/*
 * demo.c
 *
 *  Created on: 2013/05/20
 *  Copyright(C) 2013 Ryoichi Yokomizu
 *  All rights reserved
 */

#include "lcdc.h"
#include "imagebuf.h"
#include "gentype.h"
#include "sinval.h"

#define max(a, b)	((a) > (b) ? a : b)
#define min(a, b)	((a) < (b) ? a : b)

void ripple(void);
void ripple_prepare(void);
void raindrop(void);
void lineart(void);
void fin(void);

int main(int argc, char *pargv[])
{
	LCDC_Init(LCDC_MODE_RGB565);

	ripple_prepare();

	lineart();
	raindrop();
	ripple();
	fin();

	return 0;
}

static unsigned long next = 10;

int myrand(void)  /* RAND_MAX assumed to be 32767. */
{
    next = next * 1103515245 + 12345;
    return((unsigned)(next/65536) % 32768);
}

#define RIPPLE_WIDTH	480
#define RIPPLE_HEIGHT	272
#define	MAX_DIVISION	256	/* 2PI（円周）を何分割するか：精度 */
#define	MAX_WAVELENGTH	40	/* 振幅の階調 */

#define	RIPPLESTEP		2	/* 波長 */
#define	RIPPLEHEIGHT	6	/* 波紋の高さ（振幅） */
#define	RIPPLEHEIGHT2	4	/* よりリアルにする為に掛ける効果の大きさ（振幅） */

unsigned long ripple_desttbl[MAX_DIVISION][MAX_WAVELENGTH];
extern const unsigned char ripple_angltbl[RIPPLE_HEIGHT][RIPPLE_WIDTH];
extern const unsigned short ripple_centtbl[RIPPLE_HEIGHT][RIPPLE_WIDTH];

HIMAGEBUF hBufferRipple[MAX_WAVELENGTH], hBufferBmp;

void ripple_prepare(void)
{
	short rad, j, sinval, cosval, sita;
	long dget;

	// 角度に対応し、位相をずらしたテーブルを作成する。
	// それぞれの角度で、ピクセル的な距離を0～2πの範囲でsin関数によって求める。
	// -->角度（MAX:２π）の分割数はMAX_DIVISIONになる
	// -->それぞれの角度内でsin関数（MAX:２π）の分割数はMAX_WAVELENGTHになる

	for (rad = 0; rad < MAX_DIVISION; rad++) {
		sinval = mysin(rad);
		cosval = mycos(rad);
		for (j = 0; j < MAX_WAVELENGTH; j++) {
			sita = j * (SIN_VAL_NUM / MAX_WAVELENGTH);	/* sitaの値を0~2πにマッピング */
			dget = (mysin(sita) * RIPPLEHEIGHT) / SIN_VAL_MAX;	/* 単純なSIN波 */
			ripple_desttbl[rad][j] = ((cosval * dget) / SIN_VAL_MAX) * RIPPLE_WIDTH + ((sinval * dget) / SIN_VAL_MAX);
		}
	}
}

void ripple(void)
{
	HIMAGEBUF hBufferBmp = 0, hBuffer = 0;
	int count, step, x, y, idx, idy, pixelpos, ofs, maxofs;
	unsigned short *psrc, *pdest;

	unsigned char *pangltbl;
	unsigned short *pcenttbl;

	hBuffer = IMG_CreateOffscreenBuffer();

	printf("Loading bitmap...\n");
	hBufferBmp = IMG_CreateOffscreenBuffer();
	if (IMG_LoadBMP(hBufferBmp, "tahichi2.bmp") != 0) {
		printf("Cannot open 'tahichi2.bmp'.\n");
		IMG_DeleteBuffer(hBufferBmp);
		IMG_DeleteBuffer(hBuffer);
		return;
	}

	printf("Preparing table...\n");
	ripple_prepare();
	printf("Start!\n");

	LCDC_ClearScreen();
	usleep(1000000);
	LCDC_DrawString(138, 50, 1, 1, "ATM0430D5 + S1D13781 DEMONSTRATION", 255,255,255);
	LCDC_DrawString(138, 100,2, 2, "Scene3 : Ripple effect", 0,255,0);
	usleep(5000000);
	LCDC_ClearScreen();
	usleep(1000000);

	psrc = (unsigned short *)IMG_GetBits(hBufferBmp, IMG_GETBITS_READONLY);
	maxofs = RIPPLE_WIDTH * RIPPLE_HEIGHT;

	for (count = 0; count < 1000; count++) {
		pdest = (unsigned short *)IMG_GetBits(hBuffer, IMG_GETBITS_NORMAL);
		step = MAX_WAVELENGTH - count % MAX_WAVELENGTH;	// 内側→外側の波紋
		pangltbl = &ripple_angltbl[0][0];
		pcenttbl = &ripple_centtbl[0][0];
		// 仮想画面上に、テーブルから求めた波紋の表に従ってピクセルを配置する
		for (y = 0; y < RIPPLE_HEIGHT * RIPPLE_WIDTH; y++) {
				// 各テーブルからピクセルを配置する位置を決定する
				idy = *pangltbl++;
				idx = (*pcenttbl++ + step) % MAX_WAVELENGTH;
				ofs = y + ripple_desttbl[idy][idx];
				ofs = max(0, min(maxofs, ofs));	// クリッピング処理
				// バックバッファに書き込む
				*pdest++ = *(psrc + ofs);
		}
		IMG_TransferToScreen(hBuffer);
	}

	IMG_DeleteBuffer(hBufferBmp);
	IMG_DeleteBuffer(hBuffer);
}

#define LINE_NUM	40

void lineart(void)
{
	HIMAGEBUF hBuffer = 0;
	short x1[LINE_NUM], y1[LINE_NUM];
	short x2[LINE_NUM], y2[LINE_NUM];
	short dx1, dy1, dx2, dy2, dr, dg, db;
	short r[LINE_NUM], g[LINE_NUM], b[LINE_NUM];
	long i, count;

	LCDC_ClearScreen();
	usleep(1000000);
	LCDC_DrawString(138, 50, 1, 1, "ATM0430D5 + S1D13781 DEMONSTRATION", 255,255,255);
	LCDC_DrawString(138, 100,2, 2, "Scene1 : Line art", 0,255,0);
	usleep(5000000);
	LCDC_ClearScreen();
	usleep(1000000);

	hBuffer = IMG_CreateOffscreenBuffer();

	for(i = 0; i < LINE_NUM; i++) {
		x1[i] = y1[i] = -1;
	}

	x1[0] = myrand() % 50;
	y1[0] = myrand() % 50;
	x2[0] = myrand() % 50 + 430;
	y2[0] = myrand() % 50 + 222;
	r[0] = 255;
	g[0] = 255;
	b[0] = 255;
	dx1 = dy1 = 8;
	dx2 = dy2 = -6;
	dr = (myrand() % 31 - 15);
	dg = (myrand() % 31 - 15);
	db = (myrand() % 31 - 15);

	for (count = 0; count < 400; count++) {

		IMG_Fill(hBuffer, 0,0,0);
		for (i = 0; i < LINE_NUM; i++) {
			if (x1[i] > 0) {
				IMG_Line(hBuffer, x1[i],y1[i], x2[i],y2[i], (unsigned char)r[i],(unsigned char)g[i],(unsigned char)b[i]);
			}
		}
		for (i = LINE_NUM - 1; i >= 1; i--) {
			x1[i] = x1[i-1];
			y1[i] = y1[i-1];
			x2[i] = x2[i-1];
			y2[i] = y2[i-1];
			r[i] = r[i-1];
			g[i] = g[i-1];
			b[i] = b[i-1];
		}
		if ((x1[0] + dx1 < 0) || (x1[0] + dx1 > 479)) {
			dx1 = -dx1;
		}
		if ((y1[0] + dy1 < 0) || (y1[0] + dy1 > 271)) {
			dy1 = -dy1;
		}
		if ((x2[0] + dx2 < 0) || (x2[0] + dx2 > 479)) {
			dx2 = -dx2;
		}
		if ((y2[0] + dy2 < 0) || (y2[0] + dy2 > 271)) {
			dy2 = -dy2;
		}
		x1[0] += dx1;
		y1[0] += dy1;
		x2[0] += dx2;
		y2[0] += dy2;

		r[0] += dr;
		g[0] += dg;
		b[0] += db;
		if (r[0] < 0) {
			dr = myrand() % 30;
			r[0] = 0;
		}
		else if(r[0] > 255) {
			dr = -myrand() % 30;
			r[0] = 255;
		}
		else {
			;
		}
		if (g[0] < 0) {
			dg = myrand() % 30;
			g[0] = 0;
		}
		else if(g[0] > 255) {
			dg = -myrand() % 30;
			g[0] = 255;
		}
		else {
			;
		}
		if (b[0] < 0) {
			db = myrand() % 30;
			b[0] = 0;
		}
		else if(b[0] > 255) {
			db = -myrand() % 30;
			b[0] = 255;
		}
		else {
			;
		}

		IMG_TransferToScreen(hBuffer);
	}

	IMG_DeleteBuffer(hBuffer);
}

#define CIRCLE_NUM	50

void raindrop(void)
{
	HIMAGEBUF hBuffer = 0;
	short cx[CIRCLE_NUM], cy[CIRCLE_NUM], rn[CIRCLE_NUM], rn1[CIRCLE_NUM];
	unsigned char r[CIRCLE_NUM], g[CIRCLE_NUM], b[CIRCLE_NUM];
	short i;
	long count;

	LCDC_ClearScreen();
	usleep(1000000);
	LCDC_DrawString(138, 50, 1, 1, "ATM0430D5 + S1D13781 DEMONSTRATION", 255,255,255);
	LCDC_DrawString(138, 100,2, 2, "Scene2 : Rain drop", 0,255,0);
	usleep(5000000);
	LCDC_ClearScreen();
	usleep(1000000);

	for (i = 0; i < CIRCLE_NUM; i++) {
		cx[i] = cy[i] = -1;
	}

	hBuffer = IMG_CreateOffscreenBuffer();

	for (count = 0; count < 300; count++) {

		IMG_Fill(hBuffer, 0,0,0);	/* オフスクリーンバッファのクリア */
		for (i = 0; i < CIRCLE_NUM; i++) {
			if (cx[i] == -1) {
				cx[i] = myrand() % 480;		/* 円の中心点(x) */
				cy[i] = myrand() % 272;		/* 円の中心点(y)	*/
				rn[i] = 2;					/* 円の半径（初期値) */
				rn1[i] = myrand() % 60 + 20;/* 円の半径 */
				r[i] = myrand() % 223 + 32;	/* カラー(R要素) */
				g[i] = myrand() % 223 + 32;	/* カラー(G要素) */
				b[i] = myrand() % 223 + 32;	/* カラー(B要素) */
			}
			/* 円弧の描画 */
			IMG_Circle(hBuffer, cx[i], cy[i], rn[i], r[i], g[i], b[i]);
			/* 一定以上のサイズになったら削除 */
			rn[i] += 2;
			if (rn[i] > rn1[i]) {
				cx[i] = cy[i] = -1;
			}
			else {
				/* １フレームごとに輝度を3%づつ下げていく */
				r[i] = (r[i] * 97) / 100;
				g[i] = (g[i] * 97) / 100;
				b[i] = (b[i] * 97) / 100;
			}
		}
		IMG_TransferToScreen(hBuffer);
	}

	IMG_DeleteBuffer(hBuffer);
}

void fin(void)
{
	LCDC_ClearScreen();
	usleep(1000000);
	LCDC_DrawString(138, 50, 1, 1, "ATM0430D5 + S1D13781 DEMONSTRATION", 255,255,255);
	LCDC_DrawString(138, 100,3, 4, "Fin.", 0,255,0);
}
