#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>
#include <errno.h>
	
#define SPI_DEVICE    "/dev/spidev0.0"  //  SPI foCX
#define SPI_SPEED     32000000          //  NbN 32MHz
#define SPI_BITS      8                 //  rbgi8bit ̂݉\j
#define SPI_BLOCKSIZE 2048              //  ]ubÑoCg
#define SPI_DELAY     0                 //  CE0 ̒x
#define SPI_CMDBLOCK  1					//  R}hMpɎg]ubN ryk

static int spi_fd;                      //  t@CfBXNv^
static int spi_numFullBlocks,           //  ڂ̃ubN
           spi_lastBlockSize;           //  ςݎcubÑoCg
static int spi_bufsize;					//	Ɏw肵obt@TCY
static struct spi_ioc_transfer *spi_tr = NULL;      //  SPI ]̍\
//  spi_init: M摜ioCgzj̑傫w肵
//            ŏɂP񂾂Ăяo
void spi_init (unsigned int bufsize)
{
    int  ret;
    //  SPI foCXJ
    spi_fd = open(SPI_DEVICE, O_RDWR);
    if (spi_fd < 0) {
        printf("error: cannot open %s (spi_init)\n", SPI_DEVICE);
        exit(-1);
    }
    //  ][h 0 𑗐MEMɐݒiI}WiCj
    unsigned char mode = SPI_MODE_0;
    ret = ioctl(spi_fd, SPI_IOC_WR_MODE, &mode);
    if (ret < 0) {
        printf("error: cannot set WR_MODE to %d (spi_init)\n", mode);
        close(spi_fd);
        exit(-1);
    }
    ret = ioctl(spi_fd, SPI_IOC_RD_MODE, &mode);
    if (ret < 0) {
        printf("error: cannot set RD_MODE to %d (spi_init)\n", mode);
        close(spi_fd);
        exit(-1);
    }
    //  rbg𑗐MEMɐݒi8bit 󂯕tȂj
    unsigned char bits = SPI_BITS;
    ret = ioctl(spi_fd, SPI_IOC_WR_BITS_PER_WORD, &bits);
    if (ret < 0) {
        printf("error: cannot set WR_BITS_PER_WORD to %d (spi_init)\n", bits);
        close(spi_fd);
        exit(-1);
    }
    ret = ioctl(spi_fd, SPI_IOC_RD_BITS_PER_WORD, &bits);
    if (ret < 0) {
        printf("error: cannot set RD_BITS_PER_WORD to %d (spi_init)\n", bits);
        close(spi_fd);
        exit(-1);
    }
    //  ]xiNbNgj𑗐MEMɐݒ
    unsigned int speed = SPI_SPEED;
    ret = ioctl(spi_fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed);
    if (ret < 0) {
        printf("error: cannot set WR_MAX_SPEED_HZ to %d (spi_init)\n", speed);
        close(spi_fd);
        exit(-1);
    }
    ret = ioctl(spi_fd, SPI_IOC_RD_MAX_SPEED_HZ, &speed);
    if (ret < 0) {
        printf("error: cannot set RD_MAX_SPEED_HZ to %d (spi_init)\n", speed);
        close(spi_fd);
        exit(-1);
    }
    //  SPI ]\̂iKvjAP[g
    spi_numFullBlocks = bufsize / SPI_BLOCKSIZE;
    spi_lastBlockSize = bufsize % SPI_BLOCKSIZE;
    spi_tr = (struct spi_ioc_transfer *)
             malloc(sizeof(struct spi_ioc_transfer) * (spi_numFullBlocks + 1 + SPI_CMDBLOCK));
	spi_bufsize = bufsize;
}
//  spi_quit: SPI gIĂяoiVl̂݁j
void spi_quit ()
{
    close(spi_fd);
    free(spi_tr);
}
//  spi_send_buffer: oCgzi bufsize ̃f[^j̑M
void spi_send_buffer (unsigned char *buffer)
{
    //  ]ubNiځjp
    int i;
    unsigned char *bufPointer = buffer;
    for (i = 0; i < spi_numFullBlocks; i++) {
        spi_tr[i].tx_buf = (unsigned long) bufPointer;
        spi_tr[i].rx_buf = (unsigned long) NULL;
        spi_tr[i].len    = SPI_BLOCKSIZE;
        spi_tr[i].speed_hz      = SPI_SPEED;
        spi_tr[i].delay_usecs   = SPI_DELAY;
        spi_tr[i].bits_per_word = SPI_BITS;
        spi_tr[i].cs_change     = 0;
        bufPointer += SPI_BLOCKSIZE;
    }
    //  ςݎc̓]ubNǉiKvɉāj
    int numBlocks;
    if (spi_lastBlockSize) {
        //  ςݎc
        i = spi_numFullBlocks;
        spi_tr[i].tx_buf = (unsigned long) bufPointer;
        spi_tr[i].rx_buf = (unsigned long) NULL;
        spi_tr[i].len    = spi_lastBlockSize;
        spi_tr[i].speed_hz      = SPI_SPEED;
        spi_tr[i].delay_usecs   = SPI_DELAY;
        spi_tr[i].bits_per_word = SPI_BITS;
        spi_tr[i].cs_change     = 0;
        numBlocks = spi_numFullBlocks + 1;
    }
    else {
        //  ҂
        numBlocks = spi_numFullBlocks;
    }
    //  ]ubN̑M
    int ret = ioctl(spi_fd, SPI_IOC_MESSAGE(numBlocks), spi_tr);
    if (ret < 0) {
        printf("error: cannot send spi messages (spi_send_buffer)\n");
        exit(-1);
    }
}

//  LCDC̃WX^Write
void LCDC_WriteRegister16(unsigned int addr, unsigned short data)
{
	//  setup a block
	static unsigned char tdata[1+3+2];	// Command(1)+Address(3)+Data(2)
	tdata[0] = 0x88;						// 16bit Write Sequence
	tdata[1] = (addr & 0x00070000) >> 16;	// A[18:16]
	tdata[2] = (addr & 0x0000ff00) >>  8;	// A[15:8]
	tdata[3] =  addr & 0x000000ff;			// A[7:0]
	tdata[4] = (data & 0xff00) >> 8;		// D[15:8]
	tdata[5] =  data & 0x00ff;				// D[7:0]
 
	struct spi_ioc_transfer tr[1];
	tr[0].tx_buf = (unsigned int) tdata;
	tr[0].rx_buf = (unsigned int) NULL;
	tr[0].len    = sizeof(tdata);
	tr[0].speed_hz      = SPI_SPEED;
	tr[0].delay_usecs   = SPI_DELAY;
	tr[0].bits_per_word = SPI_BITS;
	tr[0].cs_change     = 0;
	//  send this byte
	int ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), tr);
	if (ret < 0) {
		printf("error: cannot send spi message (spi_send16)\n");
		exit(-1);
	}
}

/* LCDCWX^(24bit) */
void LCDC_WriteRegister24(unsigned long addr, unsigned long data)
{
	//  setup a block
	static unsigned char tdata[1+3+3];	// Command(1)+Address(3)+Data(3)
	tdata[0] = 0x80;						// 8bit Write Sequence
	tdata[1] = (addr & 0x00070000) >> 16;	// A[18:16]
	tdata[2] = (addr & 0x0000ff00) >>  8;	// A[15:8]
	tdata[3] =  addr & 0x000000ff;			// A[7:0]
	tdata[4] = (data & 0x00ff0000) >> 16;	// D[23:16]
	tdata[5] = (data & 0x0000ff00) >> 8;	// D[15:7]
	tdata[6] =  data & 0x000000ff;			// D[7:0]
 
	struct spi_ioc_transfer tr[1];
	tr[0].tx_buf = (unsigned int) tdata;
	tr[0].rx_buf = (unsigned int) NULL;
	tr[0].len    = sizeof(tdata);
	tr[0].speed_hz      = SPI_SPEED;
	tr[0].delay_usecs   = SPI_DELAY;
	tr[0].bits_per_word = SPI_BITS;
	tr[0].cs_change     = 0;
	//  send this byte
	int ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), tr);
	if (ret < 0) {
		printf("error: cannot send spi message (spi_send16)\n");
		exit(-1);
	}
}

/* LCDCWX^ǂݍ(16bit) */
unsigned short LCDC_ReadRegister16(unsigned long addr)
{
	//  setup a block
	static unsigned char tdata[1+3+2+2];	// Command(1)+Address(3)+Dummy(4)
	static unsigned char rdata[6+2];		// Dummy(6)+Data(2)
	tdata[0] = 0xC8;						// 16bit Read Sequence
	tdata[1] = (addr & 0x00070000) >> 16;	// A[18:16]
	tdata[2] = (addr & 0x0000ff00) >>  8;	// A[15:8]
	tdata[3] =  addr & 0x000000ff;			// A[7:0]
	tdata[4] = 0x00;						// Dummy
	tdata[5] = 0x00;						// Dummy
	tdata[6] = 0x00;						// Dummy
	tdata[7] = 0x00;						// Dummy
 
	struct spi_ioc_transfer tr[1];
	tr[0].tx_buf = (unsigned int) tdata;
	tr[0].rx_buf = (unsigned int) rdata;
	tr[0].len    = sizeof(tdata);
	tr[0].speed_hz      = SPI_SPEED;
	tr[0].delay_usecs   = SPI_DELAY;
	tr[0].bits_per_word = SPI_BITS;
	tr[0].cs_change     = 0;

	//  send this byte
	int ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), tr);
	if (ret < 0) {
		printf("error: cannot send spi message (spi_send16)\n");
		exit(-1);
	}
	
	return (rdata[6] << 8) | rdata[7];
}

/* LCDCWX^ǂݍ(24bit) */
unsigned long LCDC_ReadRegister24(unsigned long addr)
{
	//  setup a block
	static unsigned char tdata[1+3+5];		// Command(1)+Address(3)+Dummy(5)
	static unsigned char rdata[6+3];		// Dummy(6)+Data(3)
	tdata[0] = 0xC0;						// 16bit Read Sequence
	tdata[1] = (addr & 0x00070000) >> 16;	// A[18:16]
	tdata[2] = (addr & 0x0000ff00) >>  8;	// A[15:8]
	tdata[3] =  addr & 0x000000ff;			// A[7:0]
	tdata[4] = 0x00;						// Dummy
	tdata[5] = 0x00;						// Dummy
	tdata[6] = 0x00;						// Dummy
	tdata[7] = 0x00;						// Dummy
	tdata[8] = 0x00;						// Dummy
 
	struct spi_ioc_transfer tr[1];
	tr[0].tx_buf = (unsigned int) tdata;
	tr[0].rx_buf = (unsigned int) rdata;
	tr[0].len    = sizeof(tdata);
	tr[0].speed_hz      = SPI_SPEED;
	tr[0].delay_usecs   = SPI_DELAY;
	tr[0].bits_per_word = SPI_BITS;
	tr[0].cs_change     = 0;

	//  send this byte
	int ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), tr);
	if (ret < 0) {
		printf("error: cannot send spi message (spi_send16)\n");
		exit(-1);
	}
	
	return ((unsigned long)rdata[6] << 16) | ((unsigned long)rdata[7] << 8) | (unsigned long)rdata[8];
}

#define SPI_BUFSIZE		2048
#define SPI_CMDSIZE		4
#define SPI_IMGSIZE		(SPI_BUFSIZE - SPI_CMDSIZE)

/* \f[^] */
void LCDC_BurstWrite(unsigned long addr, unsigned char* buf, unsigned long size)
{
	static unsigned char txbuf[SPI_BUFSIZE]; /* Mf[^FCommand(4)+Data(2044) */
	struct spi_ioc_transfer spi_tr;
	int count, txsize;
	
	while (size > 0) {
		txbuf[0] = 0x80;						// 8bit Write Sequence
		txbuf[1] = (addr & 0x00070000) >> 16;	// A[18:16]
		txbuf[2] = (addr & 0x0000ff00) >>  8;	// A[15:8]
		txbuf[3] =  addr & 0x000000ff;			// A[7:0]
		
		txsize = (size > SPI_IMGSIZE) ? SPI_IMGSIZE : size;
		memcpy(txbuf + SPI_CMDSIZE, buf, txsize);
		
		spi_tr.tx_buf = (unsigned int) txbuf;
		spi_tr.rx_buf = (unsigned int) NULL;
		spi_tr.len    = SPI_CMDSIZE + txsize;
		spi_tr.speed_hz      = SPI_SPEED;
		spi_tr.delay_usecs   = SPI_DELAY;
		spi_tr.bits_per_word = SPI_BITS;
		spi_tr.cs_change     = 0;
	
		//  ]ubN̑M
		int ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), &spi_tr);
		if (ret < 0) {
		    printf("error: cannot send spi messages (LCDC_BurstWrite) errno=%d\n", errno);
		    exit(-1);
		}

		addr += txsize;
		buf += txsize;
		size -= txsize;
	}
}

/* \f[^]iXV̂݁j*/
void spi_update_buffer(unsigned long addr, unsigned char *buf, unsigned long size, unsigned char first)
{
	static unsigned long bufsize = 0, blocknum = 0;
	static unsigned char *s_bufptr = NULL;
	struct spi_ioc_transfer spi_tr;
	unsigned char *bufptr;
	int count, txsize;

	if (first) {
		blocknum = (size + SPI_IMGSIZE - 1) / SPI_IMGSIZE;
		bufsize = blocknum * SPI_BUFSIZE;
		s_bufptr = (unsigned char *)malloc(bufsize);
	}
	
	bufptr = s_bufptr;
	while (size > 0) {
		txsize = (size > SPI_IMGSIZE) ? SPI_IMGSIZE : size;
		if (memcmp(bufptr + SPI_CMDSIZE, buf, txsize) != 0) {
			*bufptr       = 0x80;						// 8bit Write Sequence
			*(bufptr + 1) = (addr & 0x00070000) >> 16;	// A[18:16]
			*(bufptr + 2) = (addr & 0x0000ff00) >>  8;	// A[15:8]
			*(bufptr + 3) =  addr & 0x000000ff;			// A[7:0]
		
			memcpy(bufptr + SPI_CMDSIZE, buf, txsize);
		
			spi_tr.tx_buf = (unsigned int) bufptr;
			spi_tr.rx_buf = (unsigned int) NULL;
			spi_tr.len    = SPI_CMDSIZE + txsize;
			spi_tr.speed_hz      = SPI_SPEED;
			spi_tr.delay_usecs   = SPI_DELAY;
			spi_tr.bits_per_word = SPI_BITS;
			spi_tr.cs_change     = 0;
		
			//  ]ubN̑M
			int ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), &spi_tr);
			if (ret < 0) {
			    printf("error: cannot send spi messages (LCDC_BurstWrite) errno=%d\n", errno);
			    exit(-1);
			}
		}
		addr += txsize;
		buf += txsize;
		size -= txsize;
		bufptr += SPI_BUFSIZE;
	}
}
